#ifndef _SerialPacket_h
#define _SerialPacket_h

/*
<kit.adams@gmail.com> developed this code. You can use and modify this code for
non-commercial purposes so long as you retain this notice. You need to contact
me if you wish to use it for commercial purposes.

This work is licensed under the Creative Commons Attribution-Noncommercial-Share Alike 3.0 
New Zealand License. To view a copy of this license, visit 
http://creativecommons.org/licenses/by-nc-sa/3.0/nz/ 
or send a letter to 
Creative Commons, 171 Second Street, Suite 300, San Francisco, California, 94105, USA.

Kit Adams.
*/


/*
CHDK <-> AVR Serial Packet Protocol

Variable length packets.

The Packet ID (PID) byte and the following length byte both contain the length and packet index, although the len byte is complemented.
All the bytes in the packet except the PID have bit 6 set.

Packet:
Byte
0     PID      //Packet Id: 2 bit index : b0llllii, llll is length:1-15, 0->Handshake packet (no Length byte), 2-15 means Data packet, 15 means 14 data bytes, length==1 => 0 data bytes
1     len      //complemented l's and i's       : ~b0llllii  i.e. bit 6 is always set.
2-16  Data     //Data : 1 to 15 6 bit bytes     : b1dddddd
17    CRC6     //MSByte of CRC of bytes 1-16 (len and data bytes)    : b1dddddd
18    CRC6     //LSByte of CRC of bytes 1-16 (len and data bytes)    : b1dddddd

Max Data bytes is 14, since len = 1 => 0 data bytes (Ack packet).
Min Data packet size 4 (0.04 s)
Max Data packet size 18 (0.18 s)

Ack Handshake packets are just 0 length with no CRC: 
0 PID   //b00001ii
1 len   //complemented l's and i's:b01110ii

Packet index only needs to be 1 or 0, if a new packet is not sent until
the last one is acknowledged. The tricky case is when the packet is received
correctly and the acknowledge is sent but not received. In this case the packet
will be sent again. The receiver attempts to detect this unnecessary resend 
by seeing the index is the same as the last successful packet
and sending an acknowledge for that index immediately.

With packet index ranging from 0 to 3 we can send the next packet before the last
is acknowledged by double buffering two packets in the receiver. The LSBit of the
index is used to index the packet buffer. The packet index cannot be transmitted
as 0 because this could result in a null PID byte, so 1 is added for transmission.
In the case where 2 packets have been sent and the 2nd is acknowledged before
the 1st, the 2nd cannot be transferred to the rx ring buffer until the 1st is resent
and acknowledged. I.e. this transfer cannot happen until the buffered packet is acknowledged
and has the correct 2 bit index.

The SendByte() function pulls data from a ringbuffer of bytes. We set the MSBit (0x80)
of a byte in the ringbuffer high to mark the start of a packet. When loading a Packet from the ring buffer, the
Packet is terminated if the ringbuffer becomes empty. A new packet will be started
automatically when the current packet is full.

The low level transaction transfers 7 bits each way. Bit 6 (0x40) indicates a PID if 0, or a len or CRC
byte if set, so only 6 bits (0x3f) are available for user data. We use the convention that
a command is always the first data byte in the packet, and will therefore have bit 7 (0x80)
set, making it easy to identify commands. However, some commands may require more than 15
bytes of data, so not every packet start byte is command.

Packets of length 1 with a (6 bit) null for the first data byte (i.e. the command byte is null)
are reserved for use as synch packets by the transmission protocol. These packets are transmitted
when devices first power up and are used to ensure packet indicies match in the transmitter and
receiver. These synch packets are not transferred to the output ring buffer by ReceiveByte() function. 

*/

#include "SerIOTypes.h"


//void serialCrc6(uint8plus *crcPtr, uint8plus data);
//void serialCrc12as6(uint16plus *crcPtr, uint8plus data);
//void serialCrc14as7(uint16plus *crcPtr, uint8plus data);

void serialCrcPacket(uint16plus *crcPtr, uint8plus data);

typedef uint16plus TCRCAccum;

#include "RingBuf.h"

//This must be less than kECDataBits.
#define PACKETPAYLOADBITS 7     
#define PACKETLENGTHBITS 4     

enum
   {
   kECBufferSize = 2,    //N.B. big enough to hold a complete shoot command

   kECDataBits          = 8,//7
   kECFixedBits         = 1,
   kECIndexBits         = 2,
   kECCRCBits           = 4,
   kECInvertBit         = 1,
   kECTotalBits         = kECDataBits+kECFixedBits+kECIndexBits+kECCRCBits+kECInvertBit,
   kECInvertBitMask     = 1<<(kECTotalBits-1), //0x2000,
   kECSerialEndBitMask  = ((long)1<<kECTotalBits),
   kECTotalRxBitsMask   = ((long)1<<kECTotalBits)-1, //0x7fff,
   kECDataBitsMask      = (1<<kECDataBits)-1,
   kECIndexBitsMask     = (1<<kECIndexBits)-1,
   kECCRCBitsMask       = (1<<kECCRCBits)-1,
   kECFixedBitMask      = ((1<<kECFixedBits)-1)<<kECDataBits,
   kECFixedBitPattern   = 1<<kECDataBits,
   kECIndexShift        = kECDataBits+kECFixedBits,
   kECCRCShift          = kECIndexShift+kECIndexBits,
   kECIndexOffset       = 1,
   };



enum
   {
   kPacketIndexTxOffset = 0, //Add 1 for transmission to avoid null PID or Len bytes.
   kPacketLenTxOffset   = 1, //Add 1 for transmission to avoid null PID
   kPacketIndexEnd = 4,  //[0..4)
   kPacketBufSizeTx = 2, //Must be power of 2
   kPacketBufSizeMaskTx = kPacketBufSizeTx-1,
   kPacketBufSizeRx = 2,
   };     

enum 
   {
   kPacketLenMask = (1<<PACKETLENGTHBITS)-1,
   kMaxPacketDataSize = kPacketLenMask-kPacketLenTxOffset,
   kPacketIndexBits = 2,
   kPacketIndexMask = 3,
   };


enum
   {
   kPacketFilled  = 0x80,
   kPacketSent    = 0x40,
   kPacketReceivedAndAckSent = 0x20,
   };

enum
   {
   kCmdSynchPacket = 0, //^@
   };   

typedef struct Packet
   {
   uint8 mFlags;
   uint8 mLen;
   uint8 mTransactions; //used to determine when to resend the packet
#ifdef _DEBUG
   uint8 mRetries;
#endif   
   int8 mIndex;
   uint8 mData[kMaxPacketDataSize];
   } Packet;

int8plus IsSynchPacket(Packet *packet);

void SetSynchPacket(Packet* packet);


INLINEFUNCTION(int8plus GetPacketLen(Packet *p))
   {
   return p->mLen;
   }

INLINEFUNCTION(int8plus IsPacketSent(Packet *packet))
   {
   return packet->mFlags & kPacketSent;
   }

INLINEFUNCTION(void SetPacketSent(Packet *packet))
   {
   packet->mFlags |= kPacketSent;
   packet->mTransactions = 0;
   }

INLINEFUNCTION(int8plus IsPacketEmpty(Packet *packet))
   {
   return packet->mLen==0;
   }
   
INLINEFUNCTION(int8plus IsPacketReceivedAndAckSent(Packet *packet))
   {
   return packet->mFlags & kPacketReceivedAndAckSent;
   }
   
INLINEFUNCTION(void SetPacketReceivedAndAckSent(Packet *packet) )
   {
   packet->mFlags |= kPacketReceivedAndAckSent;
   }

INLINEFUNCTION(void SetPacketEmpty(Packet* packet))
   {
   packet->mFlags = 0;
   packet->mLen = 0;
#ifdef _DEBUG
   packet->mRetries = 0;
#endif   
   }
int8plus TransferPacketToRingBuf(Packet* packet, RingBuf *rb);
int8plus FillPacketFromRingBuf(Packet* packet, RingBuf *rb);
INLINEFUNCTION(int8plus GetPacketIndex(Packet* packet) )
   {
   return packet->mIndex;
   }
INLINEFUNCTION(void SetPacketIndex(Packet* packet, int8 index) )
   {
   packet->mIndex = index;
   }


int8plus NeedsRetry(Packet* packet);


enum
   {
   //kECBufferSize = 2,
   kTransactionsBeforeRetry = 4,
   };


enum //Tx states
   {
   kPSInit = 0,
   kPTxPID = 0,
   kPTxLen,
   kPTxData,
   kPTxCRC,
   kPTxWaitForAck,
   kPTxSendAck,
   };

enum //Rx states
   {
   kPRxPID = 0,
   //kPTxRetry,
   kPRxLen,
   kPRxData,
   kPRxCRC,
   kPRxCRC2,
   }; 
   

enum
   {
   kPSPayloadBits = PACKETPAYLOADBITS,
   kPSPayloadBitMask = (1<<kPSPayloadBits)-1,  //Available bits in a data byte.
   kByteTypeMask = (1<<kPSPayloadBits), //0x40,
   kByteTypePID  = 0x00, //0x40,
   kByteTypeLen  = kByteTypeMask,
   kByteTypeData = kByteTypeMask,
   kPacketStartFlag = 0x80,   //Data bytes with this set in the ringbuffer are put at the start of a new packet.
   }; 


INLINEFUNCTION(int8plus IsPIDByte(Byte byte) )
   {
   return (byte & kByteTypeMask) == kByteTypePID;
   }    
INLINEFUNCTION(int8plus IsLenByte(Byte byte) )
   {
   return (byte & kByteTypeMask) == kByteTypeLen;
   }    
INLINEFUNCTION(int8plus IsDataByte(Byte byte) )
   {
   return (byte & kByteTypeMask) == kByteTypeData;
   }

INLINEFUNCTION(Byte DecodePacketLen(Byte byte) )
   {
   return ((byte>>kPacketIndexBits) & kPacketLenMask)-kPacketLenTxOffset;
   }

INLINEFUNCTION(Byte EncodePacketLen(int8plus len) )
   {
   return (len+kPacketLenTxOffset)<<kPacketIndexBits;
   }

INLINEFUNCTION(int8plus GetPacketIndexFromByte(Byte byte) )
   {
   return ((byte & kPacketIndexMask) - kPacketIndexTxOffset);
   }

enum //Values for the Packet.mFlags member
   {
   kPSFlagsSendAck      = 0x80,
   kPSFlagsRxSynched    = 0x40,  //Receiver packet index has been synched with the other end
   kPSFlagsSynchAcked    = 0x20,  //Synch packet has been acknowledged by the other end.
   kPSFlagsRxedSynchPacket =0x10  //Have received a synch packet while synched - other end restarted?
   };
   
enum
   { 
   kPSFlagsRxedSynchPacketTimeout = 3,  
   };
   
   

INLINEFUNCTION(int8plus IsOKToSendAck(uint8 txState) )
   {
   return txState == kPTxPID || txState == kPTxWaitForAck;
   }

enum TCamAvr
   {
   kCamera,
   kAvr,
   kPC,
   };

typedef struct PacketSerial
   {
#ifdef _DEBUG
   TCamAvr mCamAvr;
   const char* DevName() const
      {
      return mCamAvr == kCamera ? "Camera: " : "AVR: ";
      }
#endif

   Packet mRxECPackets[kPacketBufSizeRx]; //Indexed by the packet index (0 or 1).
   Packet mTxECPackets[kPacketBufSizeTx]; 

   int8 mTxPacketIndex;
   int8 mRxPacketIndex;
   int8 mLastSynchPacketIndex;
   int8 mNextRxPacketIndex;
   uint8 mRxPacketLen;
   int8 mAckPacketIndex;
   int8 mPreRetryIndex;

   uint8 mTxByteIndex;
   uint8 mRxByteIndex;

   uint8 mTxState;
   uint8 mRxState;

   uint8 mPreAckTxState;

   TCRCAccum mRxCRC;
   TCRCAccum mTxCRC;

   uint8 mFlags;

   //Error correction
   uint8 mIndex;
   uint8 mTxCount;
   uint8 mTxBuf[kECBufferSize];

   } PacketSerial;

int8plus ReceiveByte(PacketSerial *ps, Byte rxByte,RingBuf *rb);
Byte SendByte(PacketSerial *ps, RingBuf *txRingBuf);


INLINEFUNCTION(Packet* GetCurrentTxPacket(PacketSerial *ps) )
   {
   return &ps->mTxECPackets[ps->mTxPacketIndex & (kPacketBufSizeTx-1)];
   }
INLINEFUNCTION(Packet* GetNonCurrentTxPacket(PacketSerial *ps) )
   {
   int8plus index = ps->mTxPacketIndex;
   index++;
   return &ps->mTxECPackets[index & (kPacketBufSizeTx-1)];
   }

INLINEFUNCTION(Packet* GetTxPacket(PacketSerial *ps, int8plus index) )
   {
   return &ps->mTxECPackets[index & (kPacketBufSizeTx-1)];
   }

void SetSendAckForIndex(PacketSerial *ps, uint8 index);

INLINEFUNCTION(int8 SenderECHasTxSpace(PacketSerial *ps) )
   {
   return ps->mTxCount < kECBufferSize;
   }

INLINEFUNCTION(int8 SendECGetTxByte(PacketSerial *ps) )
   {
   return ps->mTxCount ? ps->mTxBuf[ps->mIndex] : 0;
   }

int8 SenderECPut(PacketSerial *ps,int8 txCh);

void SenderECByteSent(PacketSerial *ps, RingBuf *txRingBuf);

#ifdef _DEBUG
void PacketSerialInit(PacketSerial *ps, TCamAvr camAvr);
#else
void PacketSerialInit(PacketSerial *ps);
#endif

extern void OnPacketTransferredToRingBuf(PacketSerial *ps, Packet *packet);
extern void OnCRCError();
extern void OnPacketRetry();


#endif
