/**
 * Copyright (c) 2006, Adam Dunkels
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#if DEBUG
#define DEBUG_PRINTF(...)  printf(__VA_ARGS__)
#else
#define DEBUG_PRINTF(...)
#endif

#ifdef TEST
#include <string.h>
#include <ctype.h>
#endif

#include "tokenizer.h"
#include "stdlib.h"

static char const *ptr, *nextptr;
#define MAX_NUMLEN 6

struct keyword_token {
  const char *keyword;
  int token;
};

static ubasic_token current_token = TOKENIZER_ERROR;
static int current_line = 0;

static const struct keyword_token keywords[] = {
  {"<>",            TOKENIZER_NE},
  {"<=",            TOKENIZER_LE},
  {">=",            TOKENIZER_GE},
  {"<",             TOKENIZER_LT},
  {">",             TOKENIZER_GT},
  {"not",           TOKENIZER_LNOT},
  {"or",            TOKENIZER_LOR},
  {"and",           TOKENIZER_LAND},
  {"let",           TOKENIZER_LET},
  {"if",            TOKENIZER_IF},
  {"then",          TOKENIZER_THEN},
  {"else",          TOKENIZER_ELSE},
  {"endif",         TOKENIZER_ENDIF},
  {"for",           TOKENIZER_FOR},
  {"to",            TOKENIZER_TO},
  {"next",          TOKENIZER_NEXT},
  {"step",          TOKENIZER_STEP},
  {"do",            TOKENIZER_DO},
  {"until",         TOKENIZER_UNTIL},
  {"while",         TOKENIZER_WHILE},
  {"wend",          TOKENIZER_WEND},
  {"goto",          TOKENIZER_GOTO},
  {"gosub",         TOKENIZER_GOSUB},
  {"return",        TOKENIZER_RETURN},
  {"call",          TOKENIZER_CALL},
  {"rem",           TOKENIZER_REM},
  {"cls",           TOKENIZER_CLS},
  {"print_screen",  TOKENIZER_PRINT_SCREEN},
  {"print",         TOKENIZER_PRNT},
  {"click",         TOKENIZER_CLICK},
  {"press",         TOKENIZER_PRESS},
  {"release",       TOKENIZER_RELEASE},
  {"shot",          TOKENIZER_SHOOT}, // for compatibility
  {"shoot_movie_for", TOKENIZER_SHOOT_MOVIE},
  {"shoot_when_camera_steady",TOKENIZER_SWS},
  {"shoot_when_no_movement_for",TOKENIZER_SWNM},
  {"shoot",         TOKENIZER_SHOOT},
  {"take_photo_now",TOKENIZER_SHOOT},
  {"sleep_until",   TOKENIZER_SLEEP_UNTIL},
  {"start_time",    TOKENIZER_SLEEP_UNTIL},
  {"finish_time",   TOKENIZER_FT},
  {"end_time",      TOKENIZER_FT},
  {"sleep_for_minutes",TOKENIZER_SLEEP_M}, 
  {"sleep_for_seconds",TOKENIZER_SLEEP_S}, 
  {"sleep_for_msecs",TOKENIZER_SLEEP}, 
  {"sleep_for",      TOKENIZER_SLEEP}, 
  {"sleep",         TOKENIZER_SLEEP},
  {"upload_images_for", TOKENIZER_USBUP},
  {"wait_for_switch_press",TOKENIZER_WFSP},
  /*WARNING due to tokenizer limitation longest match must be first */
  {"get_day_seconds", TOKENIZER_GET_DAY_SECONDS},
  {"get_prop",      TOKENIZER_GET_PROP},
  {"get_tick_count",TOKENIZER_GET_TICK_COUNT}, // fudgey
  {"get_user_tv_id",TOKENIZER_GET_USER_TV_ID},
  {"get_zoom_steps",TOKENIZER_GET_ZOOM_STEPS},
  {"get_zoom",      TOKENIZER_GET_ZOOM},
  {"data_received", TOKENIZER_DR},
  {"get_free_disk_space", TOKENIZER_GET_FREE_DISK_SPACE},
  {"restore_focus",             TOKENIZER_RESTORE_FOCUS},
  {"restore_zoom",              TOKENIZER_RESTORE_ZOOM},
  {"set_aflock",                TOKENIZER_SET_AFLOCK},
  {"set_av",                    TOKENIZER_SET_AV},
  {"set_backlight",             TOKENIZER_SET_BACKLIGHT},
  {"set_iso",                   TOKENIZER_SET_ISO},
  {"set_led",                   TOKENIZER_SET_LED},
  {"set_nd_filter",             TOKENIZER_SET_ND_FILTER},
  {"nd_filter_out",             TOKENIZER_NDFO},
  {"nd_filter_off",             TOKENIZER_NDFF},
  {"nd_filter_in",              TOKENIZER_NDFI},
  {"set_prop",                  TOKENIZER_SET_PROP},
  {"set_raw",                   TOKENIZER_SET_RAW},
  {"set_tv96_direct",           TOKENIZER_SET_TV96_DIRECT},
  {"set_tv",                    TOKENIZER_SET_TV},
  {"set_zoom_speed",            TOKENIZER_SET_ZOOM_SPEED},
  {"set_zoom_rel",              TOKENIZER_SET_ZOOM_REL},
  {"set_zoom_to_step",          TOKENIZER_SET_ZOOM},
  {"set_zoom_to",               TOKENIZER_SET_ZOOM},
  {"set_zoom",                  TOKENIZER_SET_ZOOM},
  {"set_focus_step_to",         TOKENIZER_SET_FS},
  {"set_focus_to",              TOKENIZER_SET_FOCUS},
  {"set_focus",                 TOKENIZER_SET_FOCUS},
  {"save_focus",                TOKENIZER_SAVE_FOCUS},
  {"save_zoom",                 TOKENIZER_SAVE_ZOOM},
  {"number_of_images_to_capture_is",  TOKENIZER_SET_BE},
  {"set_digi_focus_ref_to",        TOKENIZER_SET_FR},
  {"shutdown",                 TOKENIZER_SHUT_DOWN},
  {"shut_down",                 TOKENIZER_SHUT_DOWN},
  {"wait_click",       TOKENIZER_WAIT_CLICK},
  {"is_key",           TOKENIZER_IS_KEY},
  {"wheel_right",      TOKENIZER_WHEEL_RIGHT},
  {"wheel_left",       TOKENIZER_WHEEL_LEFT},
  {"@title",           TOKENIZER_REM},
  {"@param",           TOKENIZER_REM},
  {"@default",         TOKENIZER_REM},
  {"md_detect_motion", TOKENIZER_MD_DETECT_MOTION},
  {"md_get_cell_diff", TOKENIZER_MD_GET_CELL_DIFF},
  {"save_stack", TOKENIZER_SS},
  {"time_lapse", TOKENIZER_TL},
  {"set_autostart",    TOKENIZER_SET_SCRIPT_AUTOSTART},
  {"set_sync",         TOKENIZER_SET_SYNC},
  {"send_data",        TOKENIZER_SD},
  {"shutter_speed",    TOKENIZER_SSPD},
  {"sync_on",          TOKENIZER_SYNC_ON},
  {"synch_on",          TOKENIZER_SYNC_ON},
  {"sync_off",         TOKENIZER_SYNC_OFF},
  {"synch_off",         TOKENIZER_SYNC_OFF},
  {"readYUV",          TOKENIZER_GET_LUMINANCE},
  {"get_shooting_mode",     TOKENIZER_GET_SHOOTING}, // synonym for get_shooting
  {"get_autostart",    TOKENIZER_SCRIPT_AUTOSTART},
  {"get_usb_power",    TOKENIZER_GET_USB_POWER},
  {"get_sync",         TOKENIZER_GET_SYNC},
  {"get_flash_mode",   TOKENIZER_GET_FLASH_MODE}, //fudgey
  {"get_focus",        TOKENIZER_GET_FOCUS},
  {"get_fov",          TOKENIZER_GET_FOV},
  {"get_focused_distance",TOKENIZER_GET_FOCUS},
  {"blink_af_led_for", TOKENIZER_AFLF},
  {"af_led_on",        TOKENIZER_AFLO},
  {"af_led_off",       TOKENIZER_AFLOFF},
  {"aperture",         TOKENIZER_AP},
  {"blue_led_on",      TOKENIZER_BLUE_ON},
  {"blue_led_off",     TOKENIZER_BLUE_OFF},
  {"exit_alt",         TOKENIZER_EXIT_ALT},
  {"change_ev_by",     TOKENIZER_CEV},
//  {"disable_main_flash",TOKENIZER_DISABLE_MAIN_FLASH},
//  {"enable_main_flash",TOKENIZER_ENABLE_MAIN_FLASH},
  {"movie_mode",         TOKENIZER_MOVIE},
  {"record_mode",        TOKENIZER_RECORD},
  {"playback_mode",      TOKENIZER_PLAYBACK},
  {"make_usb_connection",TOKENIZER_MUSBC},
  {"break_usb_connection",TOKENIZER_BUSBC},
  {"enable_usb_download",TOKENIZER_ENABLE_USB_DOWNLOAD},
  {"disable_usb_download",TOKENIZER_DISABLE_USB_DOWNLOAD},
  {"get_mode",		      TOKENIZER_GET_MODE},  //Returns 0 in recordmode, 1 in playmode fudgey
  {"get_video_button",        TOKENIZER_GET_VIDEO_BUTTON},//needed for fudgey
  {"get_shooting",            TOKENIZER_GET_SHOOTING}, //needed for fudgey
  {"playsound",              TOKENIZER_PLAY_SOUND}, 
  {"beep",                   TOKENIZER_BEEP}, 
  {"get_quality",            TOKENIZER_GET_QUALITY}, 
  {"get_resolution",         TOKENIZER_GET_RESOLUTION}, 
  {"set_quality",            TOKENIZER_SET_QUALITY},
  {"set_resolution",         TOKENIZER_SET_RESOLUTION},
  {"bracketing_done",        TOKENIZER_BRACKETING},
  {"get_focus_ref",          TOKENIZER_GET_FR},
  {"equal_step_focus_bracketing",TOKENIZER_MFB},
  {"auto_focus_all",          TOKENIZER_AF_ALL},
  {"auto_focus_first",        TOKENIZER_AF_FIRST},
  {"auto_focus_bracketing",   TOKENIZER_AFB},
  {"digiscope_bracketing",    TOKENIZER_DB},
  {"hdr_bracket_1/3_ev_steps",TOKENIZER_SSB},
  {"hdr_focus_stack",         TOKENIZER_HDR_STACK},
  {"is_this_right_camera",    TOKENIZER_CAMPOS},
  {"line_1",                  TOKENIZER_LINE_ONE},
  {"line_2",                  TOKENIZER_LINE_TWO},
  {"line_3",                  TOKENIZER_LINE_THREE},
  {"each_photo_alternating",  TOKENIZER_BA},
  {"each_photo_lighter",      TOKENIZER_BL},
  {"each_photo_darker",       TOKENIZER_BD},
  {"start_custom_timer_sequence",TOKENIZER_SCTSEQ},
  {"start_continuous_sequence",TOKENIZER_SCSEQ},
  {"end_continuous_sequence", TOKENIZER_ECSEQ},
 // {"end_upload",            TOKENIZER_END_UPLOAD},
  {"wait_until_done",         TOKENIZER_WUD},
  {"lcd_on_off",              TOKENIZER_BLCD},
  {"turn_backlight_off",      TOKENIZER_TBF},
  {"turn_backlight_on",       TOKENIZER_TBN},
  {"lock_autofocus",          TOKENIZER_LA},
  {"load_edge_overlay",       TOKENIZER_LOAD_OVERLAY},
  {"reset",                   TOKENIZER_RESET},
  {"overlay_on",              TOKENIZER_OVERLAY_ON},
  {"overlay_off",             TOKENIZER_OVERLAY_OFF},
  {"unlock_autofocus",        TOKENIZER_UA},
  {"sdm_console_off",          TOKENIZER_SDMCF},
  {"sdm_console_on",          TOKENIZER_SDMCN},
  {"sdm_console_lines",       TOKENIZER_SDMCL},
  {"sdm_console_line_length", TOKENIZER_SDMCLL},
  {"usbd",                    TOKENIZER_USBD},
  {"get_script_speed",        TOKENIZER_GET_MAX_SPT},
  {"set_script_speed",        TOKENIZER_SET_MAX_SPT},
  {"tune_unit_pulse",         TOKENIZER_TUP},
  {"use_af_led",              TOKENIZER_UAFL},
  {"use_print_led",           TOKENIZER_UPL},
  {"unit_burst",              TOKENIZER_UB},
  {"unit_pulse",              TOKENIZER_UP},
  {"faster_than",             TOKENIZER_MSS},        // force minimum shutter-speed, future feature
  {"end",                     TOKENIZER_END},
  {NULL,                      TOKENIZER_ERROR}
};

//---------------------------------------------------------------------------
static int
singlechar(void)
{
  if(*ptr == '\n') {
    return TOKENIZER_CR;
  } else if(*ptr == ',') {
    return TOKENIZER_COMMA;
  } else if(*ptr == ';') {
    return TOKENIZER_SEMICOLON;
  } else if(*ptr == '+') {
    return TOKENIZER_PLUS;
  } else if(*ptr == '-') {
    return TOKENIZER_MINUS;
  } else if(*ptr == '&') {
    return TOKENIZER_AND;
  } else if(*ptr == '|') {
    return TOKENIZER_OR;
  } else if(*ptr == '^') {
    return TOKENIZER_XOR;
  } else if(*ptr == '*') {
    return TOKENIZER_ASTR;
  } else if(*ptr == '/') {
    return TOKENIZER_SLASH;
  } else if(*ptr == '%') {
    return TOKENIZER_MOD;
  } else if(*ptr == '(') {
    return TOKENIZER_LEFTPAREN;
  } else if(*ptr == ')') {
    return TOKENIZER_RIGHTPAREN;
  } else if(*ptr == '=') {
    return TOKENIZER_EQ;
  } else if(*ptr == '.') {
    return TOKENIZER_PERIOD;
  }
  return 0;
}
//---------------------------------------------------------------------------
// return error, number,cr,string, rem or variable
// 'ptr' points to current location in script
// --------------------------------------------------------------------------
static int
get_next_token(void)
{
  struct keyword_token const *kt;
  int i;

  DEBUG_PRINTF("get_next_token(): '%s'\n", ptr);

  // eat all whitespace
  while(*ptr == ' ' || *ptr == '\t' || *ptr == '\r') ptr++;

  if(*ptr == 0) {
    return TOKENIZER_ENDOFINPUT;
  }
  
  if(isdigit(*ptr)) {
    for(i = 0; i < (MAX_NUMLEN+1); ++i) {
      if(!isdigit(ptr[i])) {
	if(i > 0) {
	  nextptr = ptr + i;
	  return TOKENIZER_NUMBER;
	} else {
	  DEBUG_PRINTF("get_next_token: error due to too short number\n");
	  return TOKENIZER_ERROR;
	}
      }
      if(!isdigit(ptr[i])) {
	DEBUG_PRINTF("get_next_token: error due to malformed number\n");
	return TOKENIZER_ERROR;
      }
    }
    DEBUG_PRINTF("get_next_token: error due to too long number\n");
    return TOKENIZER_ERROR;
  } else if(*ptr == ':') {
    // label
    nextptr = ptr;
    do {
      ++nextptr;
    } while(*nextptr != ' ' && *nextptr != '\r' && *nextptr != '\n' && *nextptr != '\t');
    return TOKENIZER_LABEL;
  } else if((i=singlechar()) != 0) {
    if (i == TOKENIZER_CR){
      // move to next line, and skip all following empty lines as well
      while (singlechar() == TOKENIZER_CR) 
      {
        current_line++;
        ptr++;
        // eat all whitespace
        while(*ptr == ' ' || *ptr == '\t' || *ptr == '\r') ptr++;
      };
      ptr--;
      
      // a CR. I hope that doesn't break anything.
    }
    nextptr = ptr + 1;
    return i;
  } else if(*ptr == '"') {
    nextptr = ptr;
    do {
      ++nextptr;
    } while(*nextptr != '"');
    ++nextptr;
    return TOKENIZER_STRING;
  } else {
    for(kt = keywords; kt->keyword != NULL; ++kt) {
      if(strncmp(ptr, kt->keyword, strlen(kt->keyword)) == 0) {
	nextptr = ptr + strlen(kt->keyword);

       if (kt->token == TOKENIZER_REM)                                              // rem, @title,@param,@default
        {
           while(*nextptr != 0 && *nextptr != '\r' && *nextptr != '\n') ++nextptr;  // get following value
        }
	return kt->token;
      }
    }
  }

  if((*ptr >= 'a' && *ptr <= 'z') || (*ptr >= 'A' && *ptr <= 'Z')) {
    nextptr = ptr + 1;

    return TOKENIZER_VARIABLE;
  }

  return TOKENIZER_ERROR;
}
//---------------------------------------------------------------------------
void
tokenizer_init(const char *program)
{
  ptr = program;
  current_line = 1;
  current_token = get_next_token();
}
//---------------------------------------------------------------------------
int
tokenizer_token(void)
{
  return current_token;
}
//---------------------------------------------------------------------------
void
tokenizer_next(void)
{

  if(tokenizer_finished()) 
  {
    return;
  }

  DEBUG_PRINTF("tokenizer_next: %p\n", nextptr);
  ptr = nextptr;
  while((*ptr == ' ') || (*ptr == '\t')) 
   {
    ++ptr;
   }
  current_token = get_next_token();
  DEBUG_PRINTF("tokenizer_next: '%s' %d\n", ptr, current_token);
  return;
}
//---------------------------------------------------------------------------
int
tokenizer_num(void)
{
  return atoi(ptr);
}
//---------------------------------------------------------------------------
void
tokenizer_string(char *dest, int len)
{
  char *string_end;
  int string_len;
  
  if(tokenizer_token() != TOKENIZER_STRING) {
    return;
  }
  string_end = strchr(ptr + 1, '"');
  if(string_end == NULL) {
    return;
  }
  string_len = string_end - ptr - 1;
  if(len < string_len) {
    string_len = len;
  }
  memcpy(dest, ptr + 1, string_len);
  dest[string_len] = 0;
}
//---------------------------------------------------------------------------
void
tokenizer_label(char *dest, int len)
{
  char *string_end;
  int string_len;
  
  if(tokenizer_token() != TOKENIZER_LABEL) {
    return;
  }
  // allow string \r, \n, tabulation and space to end labels
  string_end = strpbrk(ptr + 1, " \t\r\n");

  if(string_end == NULL) {
    return;
  }
  string_len = string_end - ptr - 1;
  if(len < string_len) {
    string_len = len;
  }
  memcpy(dest, ptr + 1, string_len);
  dest[string_len] = 0;
}
//---------------------------------------------------------------------------
void
tokenizer_error_print(void)
{
  DEBUG_PRINTF("tokenizer_error_print: '%s'\n", ptr);
}
//---------------------------------------------------------------------------
int
tokenizer_finished(void)
{
  return *ptr == 0 || current_token == TOKENIZER_ENDOFINPUT;
}
//---------------------------------------------------------------------------
int
tokenizer_variable_num(void)
{
  return *ptr - (*ptr>='a'?'a':('A'-26));
}
//---------------------------------------------------------------------------
int tokenizer_line_number(void)
{
  return current_line;
}
